# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2015 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import print_function

import os

import numpy as np
import pandas as pd

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.dialects import sqlite

import six

from .tables import Base, Element, IonizationEnergy


def get_session():
    '''Return the database session connection.'''

    dbpath = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                          "elements.db")
    engine = create_engine("sqlite:///{path:s}".format(path=dbpath),
                           echo=False)
    db_session = sessionmaker(bind=engine, autoflush=False, autocommit=False)
    return db_session()


def get_engine():
    '''Return the db engine'''

    dbpath = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                          "elements.db")
    engine = create_engine("sqlite:///{path:s}".format(path=dbpath),
                           echo=False)
    return engine


def element(ids):
    '''
    Based on the type of the `ids` identifier return either an
    :py:class:`Element <mendeleev.tables.Element>` object from the
    database, or a list of :py:class:`Element <mendeleev.tables.Element>`
    objects if the `ids` is a list or a tuple of identifiers. Valid
    identifiers for an element are: *name*, *symbol*, and
    *atomic number*.

    Args:
        ids (str): element identifier

    Raises:
        ValueError: when the identifier is not a list/tuple, int or str

    Example:
        The element can be identified by symbol

        >>> from mendeleev import element
        >>> si = element('Si')
        >>> si.atomic_number
        14

        by the atomic number

        >>> al = element(13)
        >>> al.name
        'Aluminum'

        or by the name

        >>> o = element('Oxygen')
        >>> o.symbol
        'O'

        Mutiple elements can be instantiated simultaneously through as
        combination of identifiers

        >>> c, h, o = element(['C', 'Hydrogen', 8])
        >>> print(c.name, h.name, o.name)
        Carbon Hydrogen Oxygen

    '''

    if isinstance(ids, (list, tuple)):
        return [_get_element(i) for i in ids]
    elif isinstance(ids, (six.string_types, int)):
        return _get_element(ids)
    else:
        raise ValueError("Expected a <list>, <tuple>, <str> or <int>, got: {0:s}".format(type(ids)))


def _get_element(ids):
    '''
    Return an element from the database based on the `ids` identifier passed.
    Valid identifiers for an element are: *name*, *symbol*, *atomic number*.
    '''

    session = get_session()

    if isinstance(ids, six.string_types):
        if len(ids) <= 3 and ids.lower() != "tin":
            return session.query(Element).filter(Element.symbol == str(ids)).one()
        else:
            return session.query(Element).filter(Element.name == str(ids)).one()
    elif isinstance(ids, int):
        return session.query(Element).filter(Element.atomic_number == ids).one()
    else:
        raise ValueError("Expecting a <str> or <int>, got: {0:s}".format(type(ids)))


def get_all_elements():
    'Get all elements as a list'

    session = get_session()
    elements = session.query(Element).all()
    session.close()
    return elements


def get_table(tablename, **kwargs):
    '''
    Return a table from the database as `pandas <http://pandas.pydata.org/>`_
    `DataFrame <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.html>`_

    Args:
      tablename: str
        Name of the table from the database
      kwargs:
        A dictionary of keyword arguments to pass to the `pandas.read_qsl`

    Returns:
      df: pandas.DataFrame
        Pandas DataFrame with the contents of the table

    Example:
        >>> from mendeleev import get_table
        >>> df = get_table('elements')
        >>> type(df)
        pandas.core.frame.DataFrame

    '''

    tables = ['elements', 'isotopes', 'ionicradii', 'ionizationenergies',
              'groups', 'series', 'oxidationstates']

    if tablename in tables:
        engine = get_engine()
        dframe = pd.read_sql(tablename, engine, **kwargs)
        return dframe
    else:
        raise ValueError('Table should be one of: {}'.format(", ".join(tables)))


def get_attr_for_group(attr, group=18):
    '''
    A convenience function for getting a specified attribute for all
    the memebers of a given group.

    Args:
        attr : str
            Attribute of `Element` to retrieve for all the noble gasses

    Returns:
        data : dict
            Dictionary with nobel gas atomic numbers as keys and values of the
            `attr` as values
    '''

    session = get_session()
    ngs = session.query(Element).filter(Element.group_id == group).all()
    x = np.array([ng.atomic_number for ng in ngs])
    y = np.array([getattr(ng, attr) for ng in ngs])
    session.close()
    return x, y


def ids_to_attr(ids, attr='atomic_number'):
    '''
    Convert the element ids: atomic numbers, symbols, element names or a
    combination of the above to a list of corresponding attributes.

    Args:
      ids: list, str or int
        A list of atomic number, symbols, element names of a combination of
        them
      attr: str
        Name of the desired attribute

    Returns:
      out: list
        List of attributes corresponding to the ids
    '''

    if isinstance(ids, (list, tuple)):
        return [getattr(e, attr) for e in element(ids)]
    else:
        return [getattr(element(ids), attr)]


def get_ips(ids=None, deg=1):
    '''
    Return a pandas DataFrame with ionization energies for a set of elements.

    Args:
      ids: list, str or int
        A list of atomic number, symbols, element names of a combination of the
        above. If nothing is specified all elements are selected.
      deg: int or list of int
        Degree of ionization, either as int or a list of ints. If a list is
        passed then the output will contain ionization energies corresponding
        to particalr degrees in columns.

    Returns:
      df: DataFrame
        Pandas DataFrame with atomic numbers, symbols and ionization energies
    '''

    session = get_session()
    engine = get_engine()

    if ids is None:
        atns = list(range(1, 119))
    else:
        atns = ids_to_attr(ids, attr='atomic_number')

    query = session.query(Element.atomic_number, Element.symbol).filter(Element.atomic_number.in_(atns))
    df = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()), engine)

    if isinstance(deg, (list, tuple)):
        if all(isinstance(d, int) for d in deg):
            for d in deg:
                query = session.query(IonizationEnergy).\
                            filter(IonizationEnergy.degree == d).\
                            filter(IonizationEnergy.atomic_number.in_((atns)))
                out = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()), engine)
                out = out[['atomic_number', 'energy']]
                out.columns = ['atomic_number', 'IP{0:d}'.format(d)]
                df = pd.merge(df, out, on='atomic_number', how='left')
        else:
            raise ValueError('deg should be a list of ints')
    elif isinstance(deg, int):
        query = session.query(IonizationEnergy).\
                    filter(IonizationEnergy.degree == deg).\
                    filter(IonizationEnergy.atomic_number.in_((atns)))
        out = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()), engine)
        out = out[['atomic_number', 'energy']]
        out.columns = ['atomic_number', 'IP{0:d}'.format(deg)]
        df = pd.merge(df, out, on='atomic_number', how='left')
    else:
        raise ValueError('deg should be an int or a list or tuple of ints')

    return df


def get_ionic_radii(values='ionic_radius'):
    '''
    Return a pandas DataFrame with ionic radii for a set of elements.

    Args:
      values : str
        The values to be returned either `ionic_radius` or `crystal_radius`

    Returns:
      df: DataFrame
        Pandas DataFrame with atomic numbers, symbols and ionic radii for all
        coordinations
    '''

    if values not in ['ionic_radius', 'crystal_radius']:
        raise ValueError('wrong values, should be "ionic_radius" or "crystal_radius"')

    ir = get_table('ionicradii')
    ir = ir[ir.spin != 'HS']
    # create new temporary pseudo multiindex
    ir['idx'] = ir.atomic_number.astype(str) + '(' + ir.charge.astype(str) + ')'
    df = ir.pivot(index='idx', columns='coordination', values='ionic_radius')
    # get back the atomic_number and charge columns from idx
    df['atomic_number'] = df.index.str.extract(r'^(\d+)')
    df['charge'] = df.index.str.extract(r'^\d+\((-?\d+)\)')
    df.reset_index(inplace=True)
    df.drop('idx', axis=1, inplace=True)

    return df


def deltaN(id1, id2, charge1=0, charge2=0, missingIsZero=True):
    '''
    Calculate the approximate fraction of transferred electrons between
    elements or ions `id1` and `id2` with charges `charge1` and `charge2`
    respectively according to the expression

    .. math::

       \Delta N = \\frac{\chi_{A} - \chi_{B}}{2(\eta_{A} + \eta_{B})}

    Args:
      id1: str or int
        Element identifier atomic number, symbol or element name
      id2: str or int
        Element identifier atomic number, symbol or element name
    '''

    session = get_session()
    atns = ids_to_attr([id1, id2], attr='atomic_number')

    e1, e2 = [session.query(Element).filter(Element.atomic_number == a).one()
              for a in atns]

    chi = [x.en_mulliken(charge=c, missingIsZero=missingIsZero)
           for x, c in zip([e1, e2], [charge1, charge2])]

    if all(x is not None for x in chi):
        return (chi[0] - chi[1]) / (2.0 * (e1.hardness(charge=charge1) +
                                           e2.hardness(charge=charge2)))
    else:
        return None


def attributes(elem, names, fmt='8.3f'):
    '''
    Return a list of strings of all the attributes of ``elem`` specified in
    ``names``
    '''

    return ['\t{0:s} = {1:{fmt}}'.format(name.replace('_', ' ').capitalize(),
                                         getattr(elem, name), fmt=fmt)
                                         for name in names]


def n_eff(n):
    '''Return the effective principal quantum number

    Args:
        n : (int)
            Principal quantum number

    .. note::
       The values are taken from J. A. Pople, D. L. Beveridge,
       "Approximate Molecular Orbital Theory", McGraw-Hill, 1970
    '''

    values = {1: 1.0, 2: 2.0, 3: 3.0, 4: 3.7, 5: 4.0, 6: 4.2}

    return values.get(n, None)




import urllib
import numpy as np
import pandas as pd
from bs4 import BeautifulSoup


def parse_rowspan_table(table):
    '''
    based on http://stackoverflow.com/a/28766684
    '''

    tmp = table.find_all('tr')

    first = tmp[0]
    allRows = tmp[1:-1]

    headers = [header.get_text() for header in first.find_all('th')]

    results = [[data.get_text() for data in row.find_all('td')]
               for row in allRows]

    rowspan = []

    for no, tr in enumerate(allRows):
        tmp = []
        for td_no, data in enumerate(tr.find_all('td')):
            if data.has_attr("rowspan"):
                rowspan.append((no, td_no, int(data["rowspan"]),
                                data.get_text()))

    if rowspan:
        for i in rowspan:
            # tr value of rowspan in present in 1th place in results
            for j in range(1, i[2]):
                # Add value in next tr.
                results[i[0] + j].insert(i[1], i[3])

    return pd.DataFrame(data=results, columns=headers)


def ciaaw_atomic_masses():
    '''
    Scrape the CIAAW webpage for atomic masses and return them as
    pandas DataFrame
    '''

    url = 'http://ciaaw.org/atomic-masses.htm'

    req = urllib.request.Request(url)
    page = urllib.request.urlopen(req)
    soup = BeautifulSoup(page, "html5lib")

    table = soup.find_all('table', attrs={'id': 'mytable'})[0]
    df = parse_rowspan_table(table)

    df.loc[:, 'Z'] = df['Z'].str.strip().astype(int)
    df.loc[:, 'Element'] = df['Element'].str.strip()
    df.loc[:, 'Symbol'] = df['Symbol'].str.strip('\t').astype(str)
    df.loc[:, 'Atomic mass,ma/Da'] = df['Atomic mass,ma/Da'].str.strip()

    df.loc[:, 'is_radioactive'] = df['A'].str.contains('*', regex=False)
    df.loc[:, 'A'] = df['A'].str.extract('([0-9]+)', expand=False).astype(int)

    df.loc[:, 'AM wows'] = df['Atomic mass,ma/Da'].str.replace("\s", "")
    df[['AM lead', 'AM decimals', 'AM unc']] = df['AM wows'].str.extract(r'(\d+)\.(\d+)\((\d+)\)', expand=True)
    df['Atomic Mass'] = df['AM wows'].str.extract(r'(\d+\.\d+)', expand=False)

    # fill in the data for carbon
    df.loc[df['AM unc'].isnull(), 'AM lead'] = 12
    df.loc[df['AM unc'].isnull(), 'AM unc'] = 0

    # calculate the uncertainty
    mask = df['AM decimals'].notnull()
    dec = np.power(10, -1 * df.loc[mask, 'AM decimals'].str.len().astype(float))
    df.loc[mask, 'AM uncertainty'] = dec * df.loc[mask, 'AM unc'].astype(int)

    df.loc[(df['Symbol'] == 'C') & (df['A'] == 12), 'Atomic Mass'] = 12.0e0
    df.loc[(df['Symbol'] == 'C') & (df['A'] == 12), 'AM uncertainty'] = 0.0e0
    df.loc[:, 'Atomic Mass'] = df['Atomic Mass'].astype(float)

    df.drop('AM wows', axis=1, inplace=True)
    df.drop('AM lead', axis=1, inplace=True)
    df.drop('AM decimals', axis=1, inplace=True)
    df.drop('AM unc', axis=1, inplace=True)

    return df


def ciaaw_atomic_weights():
    '''
    Scrape the CIAAW webpage for atomic weights and return them as
    pandas DataFrame
    '''

    url = 'http://ciaaw.org/atomic-weights.htm'

    table = pd.read_html(url, attrs={'id': 'mytable'})[0]
    # drop the last row
    table = table.iloc[:-1, :]

    table = table.rename(columns={'Standard Atomic Weight': 'SAW'})
    table.loc[:, 'Z'] = table['Z'].astype(int)
    table.loc[:, 'SAW'] = table['SAW'].str.replace("\s", "")
    table[['SAW decimals', 'SAW unc']] =\
        table['SAW'].str.extract(r'\d+\.(\d+)\((\d+)\)', expand=True)

    # calculate the uncertainty
    mask = table['SAW decimals'].notnull()
    dec = np.power(10, -1 * table.loc[mask, 'SAW decimals'].str.len().astype(float))
    table.loc[mask, 'SAW uncertainty'] = dec * table.loc[mask, 'SAW unc'].astype(int)

    table['Atomic Weight'] = table['SAW'].str.extract(r'(\d+\.\d+)\(',
                                                      expand=False).astype(float)

    table.drop('SAW decimals', axis=1, inplace=True)
    table.drop('SAW unc', axis=1, inplace=True)

    return table


def ciaaw_isotopic_abundances():

    url = 'http://ciaaw.org/isotopic-abundances.htm'

    req = urllib.request.Request(url)
    page = urllib.request.urlopen(req)
    soup = BeautifulSoup(page, "html5lib")

    table = soup.find_all('table', attrs={'id': 'mytable'})[0]
    ia = parse_rowspan_table(table)

    abu = 'Representative isotopic composition'
    ia.loc[:, 'A'] = ia['A'].str.extract('([0-9]+)', expand=False).astype(int)
    ia.loc[:, 'Z'] = ia['Z'].astype(int)
    ia.loc[:, abu] = ia[abu].str.replace("\s", "")
    ia.loc[:, 'abundance'] = ia[abu].str.extract(r'(\d+\.\d+)\(', expand=False)
    ia.loc[:, 'abundance'].fillna(ia[abu].str.extract(r'\[(\d+\.\d+),',
                                  expand=False), inplace=True)

    return ia


def ciaaw_monoisotopic():
    '''
    Scrape the CIAAW webpage for monoisotopic elements and return the
    data as pandas DataFrame
    '''

    url = 'http://ciaaw.org/monoisotopic-elements.htm'

    # parse monoisotopic elements
    table = pd.read_html(url, attrs={'id': 'mytable'})[0]
    # drop the last row
    table = table.iloc[:-1, :]
    # convert atomic number to integer
    table.loc[:, 'Z'] = table.loc[:, 'Z'].astype(int)
    return table



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2015 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import print_function

import argparse
import textwrap
import colorama
from pyfiglet import Figlet
from colorama import Fore
from mendeleev import element, get_table


def clielement():
    '''
    CLI for convenient printing of properties for a given element
    '''

    colorama.init(autoreset=True)

    parser = argparse.ArgumentParser()
    parser.add_argument('element',
        help='Element identifier: symbol, name or atomic number')
    args = parser.parse_args()

    try:
        args.element = int(args.element)
    except ValueError:
        pass

    e = element(args.element)

    f = Figlet('dotmatrix', justify='center')
    symbol = f.renderText(e.symbol)

    table = get_table('elements')
    et = table[table['symbol'] == e.symbol].transpose()
    et.drop('description', inplace=True)
    et.drop('sources', inplace=True)
    et.drop('uses', inplace=True)

    et.index = et.index.str.replace('_', ' ').str.capitalize()
    et.sort_index(inplace=True)

    # print the data

    print(Fore.RED + symbol)

    if e.description is not None:
        print(Fore.BLUE + 'Description\n===========\n')
        print('\n'.join(['  ' + s for s in textwrap.wrap(e.description, 70)]))

    if e.sources is not None:
        print(Fore.BLUE + '\nSources\n=======\n')
        print('\n'.join(['  ' + s for s in textwrap.wrap(e.sources, 70)]))

    if e.uses is not None:
        print(Fore.BLUE + '\nUses\n====\n')
        print('\n'.join(['  ' + s for s in textwrap.wrap(e.uses, 70)]))

    print(Fore.GREEN + '\nProperties\n==========\n')
    print(et.to_string(justify='left', header=False))


if __name__ == "__main__":

    clielement()



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2015 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import seaborn as sns
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
from bokeh.plotting import figure, show, output_file
from bokeh.models import HoverTool, ColumnDataSource, FixedTicker
from collections import OrderedDict
from .mendeleev import get_table


def heatmap(prop, style='whitegrid', figsize=(16, 10), cmap='RdBu_r', lw=1,
            output=None, **kwargs):
    '''
    Plot a heatmap of the given property

    Args:
      prop : str
        Name of the attribute of Element object that is available from the
        elements table
      style : str
        Seaborn style option, default='whitegrid
      figsize : tuple
        Size of the plot, default=(16, 10)
      cmap : str
        Colormap to use, default='RdBu_r'
      lw : int
        Seaborn heatmap linewidths argumentm default=1,
        see http://stanford.edu/~mwaskom/software/seaborn/generated/seaborn.heatmap.html
      output : str
        File name to save the plot, by default nothing is saved
    '''

    ptable = get_table('elements')

    # add lanthanides and actinides

    keys = ['period', 'group_id', prop]
    els = ptable[keys].dropna()
    elements_rect = els.pivot(*keys)

    sns.set(font_scale=1.5, style=style, rc={"figure.figsize": figsize})
    mask = np.asarray(elements_rect.isnull())
    ax = sns.heatmap(elements_rect, cmap=cmap, mask=mask, linewidths=lw,
                     **kwargs)
    n = len(ax.xaxis.get_ticklabels())
    ax.set_yticklabels(elements_rect.index[::-1], rotation=0)
    ax.set_xticklabels(list(range(1, n + 1)))
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')
    ax.set_xlabel('Group')
    ax.set_ylabel('Period')
    if output is not None:
        plt.savefig(output)
    return ax


def colormap_column(df, column, cmap='RdBu_r', missing='#ffffff'):
    '''
    Return a new DataFrame with the same size (and index) as `df` with a column
    `cmap` containing HEX color mapping from `cmap` colormap.

    Args:
      df : DataFrmae
        Pandas DataFrame with the data
      column : str
        Name of the column to be color mapped
      cmap : str
        Name of the colormap, see matplotlib.org
      missing : str
        HEX color for the missing values (NaN or None)
    '''

    colormap = plt.get_cmap(cmap)
    cnorm = colors.Normalize(vmin=df[column].min(), vmax=df[column].max())
    scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=colormap)
    out = pd.DataFrame(index=df.index)
    mask = df[column].isnull()
    rgba = scalarmap.to_rgba(df[column])
    out.loc[:, 'cmap'] = [colors.rgb2hex(row) for row in rgba]
    out.loc[mask, 'cmap'] = missing

    return out


def periodic_plot(df, attribute='atomic_weight', title='Periodic Table',
                  width=1000, height=800, missing='#ffffff', decimals=3,
                  colorby=None, output=None, cmap='RdBu_r',
                  showfblock=True, long_version=False):
    '''
    Use Bokeh to plot the periodic table data contained in the `df`

    Args:
      df : DataFrame
        Pandas DataFrame with the data on elements
      attribute : str
        Name of the attribute to be displayed
      title : str
        Title to appear above the periodic table
      colorby : str
        Name of the column containig the colors
      width : int
        Width of the figure in pixels
      height : int
        Height of the figure in pixels
      decimals : int
        Number of decimals to be displayed in the bottom row of each cell
      missing : str
        Hex code of the color to be used for the missing values
      output : str
        Name of the output file to store the plot, should end in .html
      cmap : str
        Colormap to use, see matplotlib colormaps
      long_version : bool
        Show the long version of the periodic table with the f block between
        the s and d blocks
      showfblock : bool
        Show the elements from the f block
    '''

    ac = 'attribute_column'

    elements = df.copy()

    # calculate x and y of the main group/row elements

    elements.loc[elements['group_id'].notnull(), 'x'] = \
        elements.loc[elements['group_id'].notnull(), 'group_id'].astype(int)
    elements.loc[elements['period'].notnull(), 'y'] = \
        elements.loc[elements['period'].notnull(), 'period'].astype(int)

    if showfblock:
        if long_version:
            elements.loc[elements['x'] > 2, 'x'] = \
                elements.loc[elements['x'] > 2, 'x'] + 14
            for period in [6, 7]:
                mask = (elements['block'] == 'f') & (elements['period'] == period)
                elements.loc[mask, 'x'] = elements.loc[mask, 'atomic_number'] -\
                                             elements.loc[mask, 'atomic_number'].min() + 3
                elements.loc[mask, 'y'] = period
        else:
            for period in [6, 7]:
                mask = (elements['block'] == 'f') & (elements['period'] == period)
                elements.loc[mask, 'x'] = elements.loc[mask, 'atomic_number'] -\
                                             elements.loc[mask, 'atomic_number'].min() + 3
                elements.loc[mask, 'y'] = elements.loc[mask, 'period'] + 2.5

    # additional columns for positioning of the text

    elements.loc[:, 'y_anumber'] = elements['y'] - 0.3
    elements.loc[:, 'y_name'] = elements['y'] + 0.2
    if property:
        elements.loc[elements[attribute].notnull(), 'y_prop'] = elements.loc[elements[attribute].notnull(), 'y'] + 0.35
    else:
        elements.loc[:, 'y_prop'] = elements['y'] + 0.35

    if colorby == 'attribute':
        temp = colormap_column(elements, attribute, cmap=cmap, missing=missing)
        elements = pd.merge(elements, temp, left_index=True, right_index=True)
        colorby = 'cmap'

    if elements[attribute].dtype == np.float64:
        elements[ac] = elements[attribute].round(decimals=decimals)
    else:
        elements[ac] = elements[attribute]

    if colorby not in elements.columns:
        series = get_table('series')
        elements = pd.merge(elements, series, left_on='series_id',
                            right_on='id', suffixes=('', '_series'))
        colorby = 'color'

    source = ColumnDataSource(data=elements)

    TOOLS = "hover,save,reset"

    p = figure(title=title,
               tools=TOOLS,
               x_axis_location='above',
               x_range=(elements.x.min() - 0.5, elements.x.max() + 0.5),
               y_range=(elements.y.max() + 0.5, elements.y.min() - 0.5),
               width=width, height=height,
               toolbar_location='above',
               toolbar_sticky=False,
               )

    p.rect("x", "y", 0.9, 0.9, source=source, color=colorby, fill_alpha=0.6)

    # adjust the ticks and axis bounds
    p.yaxis.bounds = (1, 7)
    p.axis[1].ticker.num_minor_ticks = 0
    if long_version:
        # Turn off tick labels
        p.axis[0].major_label_text_font_size = '0pt'
        # Turn off tick marks
        p.axis[0].major_tick_line_color = None  # turn off major ticks
        p.axis[0].ticker.num_minor_ticks = 0  # turn off minor ticks
    else:
        p.axis[0].ticker = FixedTicker(ticks=list(range(1, 19)))

    text_props = {
        "source": source,
        "angle": 0,
        "color": "black",
        "text_align": "center",
        "text_baseline": "middle"
    }

    p.text(x="x", y="y", text="symbol",
           text_font_style="bold", text_font_size="15pt", **text_props)

    p.text(x="x", y="y_anumber", text="atomic_number",
           text_font_size="9pt", **text_props)

    p.text(x="x", y="y_name", text="name",
           text_font_size="6pt", **text_props)

    p.text(x="x", y="y_prop", text=ac,
           text_font_size="7pt", **text_props)

    p.grid.grid_line_color = None

    hover = p.select(dict(type=HoverTool))
    hover.tooltips = OrderedDict([
        ("name", "@name"),
        ("atomic number", "@atomic_number"),
        ("atomic weight", "@atomic_weight"),
        ('EN Pauling', '@en_pauling'),
        ('Electron affinity', '@electron_affinity'),
        ("CPK color", "$color[hex, swatch]:cpk_color"),
        ("Jmol color", "$color[hex, swatch]:jmol_color"),
        ("Molcas GV color", "$color[hex, swatch]:molcas_gv_color"),
        ("electronic configuration", "@electronic_configuration"),
    ])

    if output:
            output_file(output)

    show(p)




from __future__ import print_function

import numpy as np
import pandas as pd
from sqlalchemy.dialects import sqlite

from mendeleev import (element, get_table, get_engine, get_session,
                       get_attr_for_group)

from .tables import IonizationEnergy


def get_zeff(an, method='slater'):
    'A helper function to calculate the effective nuclear charge'

    e = element(an)
    return e.zeff(method=method)


def get_neutral_data():
    '''
    Get extensive set of data from multiple database tables as pandas.DataFrame
    '''

    elements = get_table('elements')
    series = get_table('series')
    groups = get_table('groups')

    elements = pd.merge(elements, series, left_on='series_id', right_on='id',
                        how='left', suffixes=('', '_series'))
    elements = pd.merge(elements, groups, left_on='group_id',
                        right_on='group_id', how='left',
                        suffixes=('', '_group'))

    elements.rename(columns={'color': 'series_colors'}, inplace=True)

    en_scales = ['allred-rochow', 'cottrell-sutton', 'gordy',
                 'martynov-batsanov', 'mulliken', 'nagle', 'sanderson']

    for scale in en_scales:
        elements['en_' + scale] = [element(row.symbol).electronegativity(scale=scale)
                                   for i, row in elements.iterrows()]

    for attr in ['hardness', 'softness']:
        elements[attr] = [getattr(element(row.symbol), attr)()
                          for i, row in elements.iterrows()]

    elements['mass'] = [element(row.symbol).mass_str()
                        for i, row in elements.iterrows()]

    elements.loc[:, 'zeff_slater'] = elements.apply(
        lambda x: get_zeff(x['atomic_number'], method='slater'), axis=1)
    elements.loc[:, 'zeff_clementi'] = elements.apply(
        lambda x: get_zeff(x['atomic_number'], method='clementi'), axis=1)

    session = get_session()
    engine = get_engine()

    query = session.query(IonizationEnergy).\
        filter(IonizationEnergy.degree == 1).\
        filter(IonizationEnergy.atomic_number.in_(list(range(1, 119))))
    out = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()),
                            engine)
    out = out[['atomic_number', 'energy']]
    out.columns = ['atomic_number', 'ionization_energy']
    elements = pd.merge(elements, out, on='atomic_number', how='left')

    return elements


def add_plot_columns(elements):
    '''
    Add columns needed for the creating the plots

    Args:
        elements: pd.DataFrame
    '''

    mask = elements['group_id'].notnull()

    elements.loc[mask, 'x'] = elements.loc[mask, 'group_id'].astype(int)
    elements.loc[:, 'y'] = elements.loc[:, 'period'].astype(int)

    elements.loc[mask, 'group_name'] = elements.loc[mask, 'group_id'].astype(int).astype(str)
    elements.loc[~mask, 'group_name'] = 'f block'

    for period in [6, 7]:
        mask = (elements['block'] == 'f') & (elements['period'] == period)
        elements.loc[mask, 'x'] = elements.loc[mask, 'atomic_number'] -\
                                        elements.loc[mask, 'atomic_number'].min() + 3
        elements.loc[mask, 'y'] = elements.loc[mask, 'period'] + 2.5

    # additional columns for positioning of the text

    elements.loc[:, 'y_symbol'] = elements['y'] - 0.05
    elements.loc[:, 'y_anumber'] = elements['y'] - 0.3
    elements.loc[:, 'y_name'] = elements['y'] + 0.18

    return elements


def get_app_data():
    'write a file with the neutral data'

    import mendeleev
    version = mendeleev.__version__

    data = get_neutral_data()
    data = add_plot_columns(data)
    fname = 'neutral_{0:s}.pkl'.format(version)
    data.to_pickle(fname)
    print('wrote file: ', fname)


def estimate(x, attribute, group=18, deg=1, kind='linear'):
    '''
    Evaluate a value `attribute` for `x` by interpolation or
    extrapolation of the data points in `data`.

    Args:
        x : int
            Value for which the property will be evaluated
        attribute : int
            Attribute to be estimated
        group : int
            Periodic table group number
        deg : int
            Degree of the polynomial used in the extrapolation beyond
            the provided data points
        kind : str
            Kind of the interpolation used, see docs for
            `numpy.interp1d`
    '''

    xref, yref = get_attr_for_group(attribute, group=group)

    if xref.min() <= x <= xref.max():
        return np.interp([x], xref, yref)
    else:
        if x < xref.min():
            xslice = xref[:3]
            yslice = yref[:3]
        elif x > xref.max():
            xslice = xref[-3:]
            yslice = yref[-3:]

        fit = np.polyfit(xslice, yslice, deg)
        fn = np.poly1d(fit)
        return fn(x)



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2015 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''tables module specifying the database model'''

import math
from operator import attrgetter
import numpy as np

from sqlalchemy import Column, Boolean, Integer, String, Float, ForeignKey
from sqlalchemy.orm import relationship, reconstructor
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method

from .econf import ElectronicConfiguration, get_l, ORBITALS


__all__ = ['Element', 'IonizationEnergy', 'IonicRadius', 'OxidationState',
           'Isotope', 'Series', 'ScreeningConstant']


Base = declarative_base()


class Element(Base):
    '''
    Chemical element.

    Attributes:
      abundance_crust : float
        Abundance in the earth's crust in mg/kg
      abundance_sea : float
        Abundance in the seas in mg/L
      annotation : str
        Annotations regarding the data
      atomic_number : int
        Atomic number
      atomic_radius : float
        Atomic radius in pm
      atomic_radius_rahm : float
        Atomic radius by Rahm et al. in pm
      atomic_volume : float
        Atomic volume in cm3/mol
      atomic_weight : float
        Relative atomic weight as the ratio of the average mass of atoms
        of the element to 1/12 of the mass of an atom of 12C
      block : int
        Block in periodic table, s, p, d, f
      boiling_point : float
        Boiling temperature in K
      c6 : float
        C_6 dispersion coefficient in a.u. from X. Chu & A. Dalgarno, J. Chem. Phys.,
        121(9), 4083–4088 (2004) doi:10.1063/1.1779576, and the value for
        Hydrogen was taken from K. T. Tang, J. M. Norbeck and P. R. Certain,
        J. Chem. Phys. 64, 3063 (1976), doi:10.1063/1.432569
      c6_gb : float
        C_6 dispersion coefficient in a.u. from Gould, T., & Bučko, T. (2016).
        JCTC, 12(8), 3603–3613. http://doi.org/10.1021/acs.jctc.6b00361
      cas : str
        Chemical Abstracts Service identifier
      covalent_radius_bragg : float
        Covalent radius in pm from
      covalent_radius_cordero : float
        Covalent radius in pm from Cordero, B., Gómez, V., Platero-Prats, A.
        E., Revés, M., Echeverría, J., Cremades, E., … Alvarez, S. (2008).
        Covalent radii revisited. Dalton Transactions, (21), 2832.
        doi:10.1039/b801115j
      covalent_radius_pyykko : float
        Single bond covalent radius in pm Pyykkö, P., & Atsumi, M. (2009).
        Molecular Single-Bond Covalent Radii for Elements 1-118.
        Chemistry - A European Journal, 15(1), 186–197.
        doi:10.1002/chem.200800987
      covalent_radius_pyykko_double : float
        Double bond covalent radius in pm from P. Pyykkö et al.
      covalent_radius_pyykko_triple : float
        Triple bond covalent radius in pm from P. Pyykkö et al.
      covalent_radius_slater : float
        Covalent radius in pm from Slater
      cpk_color : str
        CPK color of the atom in HEX,
        see http://jmol.sourceforge.net/jscolors/#color_U
      density : float
        Density at 295K in g/cm3
      description : str
        Short description of the element
      dipole_polarizability : float
        Dipole polarizability in atomic units from P. Schwerdtfeger "Table of
        experimental and calculated static dipole polarizabilities for the
        electronic ground states of the neutral elements (in atomic units)",
        February 11, 2014
      discoverers: str
        The discoverers of the element
      discovery_location: str
        The location where the element was discovered
      discovery_year: int
        The year the element was discovered
      electron_affinity : float
        Electron affinity in eV
      en_allen : float
        Allen's scale of electronegativity (Configurational energy)
      en_ghosh : float
        Ghosh's scale of enectronegativity
      en_pauling : float
        Pauling's scale of electronegativity
      econf : str
        Ground state electron configuration
      evaporation_heat : float
        Evaporation heat in kJ/mol
      fusion_heat : float
        Fusion heat in kJ/mol
      gas_basicity : Float
        Gas basicity
      geochemical_class : String
        Geochemical classification of the elements
      goldschmidt_class : String
        Goldschmidt classification of the elements
      group : int
        Group in periodic table
      heat_of_formation : float
        Heat of formation in kJ/mol
      is_monoisotopic : bool
        A flag marking if the element is monoisotopic
      jmol_color : str
        Color of the atom as used in Jmol, in HEX,
        see http://jmol.sourceforge.net/jscolors/#color_U
      lattice_constant : float
        Lattice constant in ang
      lattice_structure : str
        Lattice structure code
      mass : float
        Relative atomic mass. Ratio of the average mass of atoms
        of the element to 1/12 of the mass of an atom of 12C
      melting_point : float
        Melting temperature in K
      metallic_radius : Float
        Single-bond metallic radius or metallic radius, have been
        calculated by Pauling using interatomic distances and an
        equation relating such distances with bond number
      metallic_radius_c12 : Float
        Metallic radius obtained by Pauling with an assumed number of
        nearest neighbors equal to 12
      molcas_gv_color : str
        Color of an atom in HEX from MOLCAS GV http://www.molcas.org/GV/
      name : str
        Name in English
      name_origin: str
        Origin of the name
      period : int
        Period in periodic table
      proton_affinity : Float
        Proton affinity
      series : int
        Index to chemical series
      sources: str
        Sources of the element
      specific_heat : float
        Specific heat in J/g mol @ 20 C
      symbol : str of length 1 or 2
        Chemical symbol
      thermal_conductivity : float
        Thermal conductivity in @/m K @25 C
      uses: str
        Uses of the element
      vdw_radius : float
        Van der Waals radius in pm from W. M. Haynes, Handbook of Chemistry and
        Physics 95th Edition, CRC Press, New York, 2014, ISBN-10: 1482208679,
        ISBN-13: 978-1482208672.
      vdw_radius_bondi : float
        Van der Waals radius according to Bondi in pm
      vdw_radius_truhlar : float
        Van der Waals radius according to Truhlar in pm
      vdw_radius_rt : float
        Van der Waals radius according to Rowland and Taylor in pm
      vdw_radius_batsanov : float
        Van der Waals radius according to Batsanov in pm
      vdw_radius_dreiding : float
        Van der Waals radius from the DREIDING force field in pm
      vdw_radius_uff : float
        Van der Waals radius from the UFF in pm
      vdw_radius_mm3 : float
        Van der Waals radius from MM3 in pm
      oxistates : list
        Oxidation states
      ionenergy : dict
        Ionization energies in eV parsed from
        http://physics.nist.gov/cgi-bin/ASD/ie.pl on April 13, 2015
    '''

    __tablename__ = 'elements'

    abundance_crust = Column(Float)
    abundance_sea = Column(Float)
    annotation = Column(String)
    atomic_number = Column(Integer, primary_key=True)
    atomic_radius = Column(Float)
    atomic_radius_rahm = Column(Float)
    atomic_volume = Column(Float)
    atomic_weight = Column(Float)
    atomic_weight_uncertainty = Column(Float)
    block = Column(String)
    boiling_point = Column(Float)
    cas = Column(String)
    covalent_radius_bragg = Column(Float)
    covalent_radius_cordero = Column(Float)
    covalent_radius_pyykko = Column(Float)
    covalent_radius_pyykko_double = Column(Float)
    covalent_radius_pyykko_triple = Column(Float)
    covalent_radius_slater = Column(Float)
    c6 = Column(Float)
    c6_gb = Column(Float)
    cpk_color = Column(String)
    density = Column(Float)
    description = Column(String)
    dipole_polarizability = Column(Float)
    discoverers = Column(String)
    discovery_location = Column(String)
    discovery_year = Column(Integer)
    electron_affinity = Column(Float)
    en_allen = Column(Float)
    en_ghosh = Column(Float)
    en_pauling = Column(Float)
    econf = Column('electronic_configuration', String)
    evaporation_heat = Column(Float)
    fusion_heat = Column(Float)
    gas_basicity = Column(Float)
    geochemical_class = Column(String)
    goldschmidt_class = Column(String)
    group_id = Column(Integer, ForeignKey("groups.group_id"))
    group = relationship("Group", uselist=False, lazy='subquery')
    heat_of_formation = Column(Float)
    is_monoisotopic = Column(Boolean)
    is_radioactive = Column(Boolean)
    jmol_color = Column(String)
    lattice_constant = Column(Float)
    lattice_structure = Column(String)
    melting_point = Column(Float)
    metallic_radius = Column(Float)
    metallic_radius_c12 = Column(Float)
    molcas_gv_color = Column(String)
    name = Column(String)
    name_origin = Column(String)
    period = Column(Integer)
    proton_affinity = Column(Float)
    _series_id = Column("series_id", Integer, ForeignKey("series.id"))
    _series = relationship("Series", uselist=False, lazy='subquery')
    series = association_proxy("_series", "name")
    sources = Column(String)
    specific_heat = Column(Float)
    symbol = Column(String)
    thermal_conductivity = Column(Float)
    uses = Column(String)
    vdw_radius = Column(Float)
    vdw_radius_alvarez = Column(Float)
    vdw_radius_bondi = Column(Float)
    vdw_radius_truhlar = Column(Float)
    vdw_radius_rt = Column(Float)
    vdw_radius_batsanov = Column(Float)
    vdw_radius_dreiding = Column(Float)
    vdw_radius_uff = Column(Float)
    vdw_radius_mm3 = Column(Float)

    ionic_radii = relationship("IonicRadius", lazy='subquery')
    _ionization_energies = relationship("IonizationEnergy", lazy='subquery')
    _oxidation_states = relationship("OxidationState", lazy='subquery')
    isotopes = relationship("Isotope", lazy='subquery')
    screening_constants = relationship('ScreeningConstant', lazy='subquery')

    @reconstructor
    def init_on_load(self):
        'Initialize the ElectronicConfiguration class as attribute of self'

        self.ec = ElectronicConfiguration(self.econf)

    @hybrid_property
    def ionenergies(self):
        '''
        Return a dict with ionization degree as keys and ionization energies
        in eV as values.
        '''

        return {ie.degree: ie.energy for ie in self._ionization_energies}

    @hybrid_property
    def oxistates(self):
        '''Return the oxidation states as a list of ints'''

        return [os.oxidation_state for os in self._oxidation_states]

    @hybrid_property
    def sconst(self):
        '''
        Return a dict with screening constants with tuples (n, s) as keys and
        screening constants as values'''

        return {(x.n, x.s): x.screening for x in self.screening_constants}

    @hybrid_property
    def electrons(self):
        '''Return the number of electrons.'''

        return self.atomic_number

    @hybrid_property
    def neutrons(self):
        '''
        Return the number of neutrons of the most abundant natural stable
        isotope.
        '''

        return self.mass_number - self.protons

    @hybrid_property
    def protons(self):
        '''Return the number of protons.'''

        return self.atomic_number

    @hybrid_property
    def mass(self):
        '''
        Return the `atomic_weight` if defined or mass number otherwise.
        '''

        return self.atomic_weight

    @hybrid_property
    def mass_number(self):
        '''
        Return the mass number of the most abundant natural stable isotope
        '''

        if len(self.isotopes) > 0:
            lwithabu = [i for i in self.isotopes if i.abundance is not None]
            if len(lwithabu) > 0:
                return max(lwithabu, key=attrgetter('abundance')).mass_number
            else:
                return self.isotopes[0].mass_number
        else:
            return int(self.atomic_weight)

    def mass_str(self):
        '''String representation of atomic weight'''

        if self.atomic_weight_uncertainty is None:
            if self.is_radioactive:
                return '[{aw:.0f}]'.format(aw=self.atomic_weight)
            else:
                return '{aw:.3f}'.format(aw=self.atomic_weight)
        else:
            dec = np.abs(np.floor(np.log10(np.abs(self.atomic_weight_uncertainty)))).astype(int)
            if dec > 5:
                dec = 5
            if self.is_radioactive:
                return '[{aw:.{dec}f}]'.format(aw=self.atomic_weight, dec=dec)
            else:
                return '{aw:.{dec}f}'.format(aw=self.atomic_weight, dec=dec)

    @hybrid_property
    def covalent_radius(self):
        '''
        Return the default covalent radius which is covalent_radius_pyykko
        '''

        return self.covalent_radius_pyykko

    @hybrid_method
    def hardness(self, charge=0):
        '''
        Return the absolute hardness, calculated as

        .. math::

           \eta = \\frac{I - A}{2}

        where I is the ionization energy and A is the electron affinity

        Args:
          charge: int
            Charge of the cation for which the hardness will be calculated
        '''

        if charge == 0:
            if self.ionenergies.get(1, None) is not None and self.electron_affinity is not None:
                return (self.ionenergies[1] - self.electron_affinity) * 0.5
            else:
                return None
        elif charge > 0:
            if self.ionenergies.get(charge + 1, None) is not None and\
               self.ionenergies.get(charge, None) is not None:
                return (self.ionenergies[charge + 1] - self.ionenergies[charge]) * 0.5
            else:
                return None
        elif charge < 0:
            raise ValueError('Charge has to be a non-negative integer, got: {}'.format(charge))

    @hybrid_method
    def softness(self, charge=0):
        '''
        Return the absolute softness, calculated as

        .. math::

           S = \\frac{1}{2\eta}

        where :math:`\eta` is the absolute hardness

        Args:
          charge: int
            Charge of the cation for which the hardness will be calculated
        '''

        eta = self.hardness(charge=charge)

        if eta is None:
            return None
        else:
            return 1.0 / (2.0 * eta)

    def zeff(self, n=None, o=None, method='slater', alle=False):
        '''
        Return the effective nuclear charge for ``(n, s)``

        Args:
          method : str
            Method to calculate the screening constant, the choices are
              - `slater`, for Slater's method as in Slater, J. C. (1930).
                Atomic Shielding Constants. Physical Review, 36(1), 57–64.
                `doi:10.1103/PhysRev.36.57 <http://www.dx.doi.org/10.1103/PhysRev.36.57>`_
              - `clementi` for values of screening constants from Clementi, E.,
                & Raimondi, D. L. (1963). Atomic Screening Constants from SCF
                Functions. The Journal of Chemical Physics, 38(11), 2686.
                `doi:10.1063/1.1733573 <http://www.dx.doi.org/10.1063/1.1733573>`_
                and Clementi, E. (1967). Atomic Screening Constants from SCF
                Functions. II. Atoms with 37 to 86 Electrons. The Journal of
                Chemical Physics, 47(4), 1300.
                `doi:10.1063/1.1712084 <http://www.dx.doi.org/10.1063/1.1712084>`_
          n : int
            Principal quantum number
          o : str
            Orbital label, (s, p, d, ...)
          alle : bool
            Use all the valence electrons, i.e. calculate screening for an
            extra electron when method='slater', if method='clementi' this
            option is ignored
        '''

        # identify th valence s,p vs d,f
        if n is None:
            n = self.ec.max_n()
        else:
            if not isinstance(n, int):
                raise ValueError('<n> should be an integer, got: {}'.format(type(n)))

        if o is None:
            # take the shell with max `l` for a given `n`
            o = ORBITALS[max([get_l(x[1])
                         for x in self.ec.conf.keys() if x[0] == n])]
        else:
            if o not in ORBITALS:
                raise ValueError('<s> should be one of {}'.format(", ".join(ORBITALS)))

        if method.lower() == 'slater':
            return self.atomic_number - self.ec.slater_screening(n=n, o=o,
                                                                 alle=alle)
        elif method.lower() == 'clementi':
            sc = self.sconst.get((n, o), None)
            if sc is not None:
                return self.atomic_number - self.sconst.get((n, o), None)
            else:
                return sc
        else:
            raise ValueError('<method> should be one of {}'.format("slater, clementi"))

    def electronegativity(self, scale='pauling', charge=0):
        '''
        Calculate the electronegativity using one of the methods

        Args:
          scale : str
           Name of the electronegativity scale, one of

           - `allen`
           - `allred-rochow`
           - `cottrell-sutton`
           - `gordy`
           - `li-xue`
           - `martynov-batsanov`
           - `mulliken`
           - `nagle`
           - `pauling`
           - `sanderson`
        '''

        # TODO:
        #        add an option to convert the value to Pauling units
        #        pu : bool
        #            Convert to Pauling's units

        if scale == 'allen':
            return self.en_allen
        elif scale == 'allred-rochow':
            return self.zeff(alle=True) / math.pow(self.covalent_radius, 2)
        elif scale == 'cottrell-sutton':
            return math.sqrt(self.zeff(alle=True) / self.covalent_radius)
        elif scale == 'gordy':
            return self.zeff(alle=True) / self.covalent_radius
        elif scale == 'li-xue':
            return self.en_li_xue(charge=charge)
        elif scale == 'martynov-batsanov':
            return self.en_martynov_batsanov()
        elif scale == 'mulliken':
            return self.en_mulliken()
        elif scale == 'nagle':
            if self.dipole_polarizability is not None:
                return math.pow(self.nvalence() / self.dipole_polarizability, 1.0 / 3.0)
            else:
                return None
        elif scale == 'pauling':
            return self.en_pauling
        elif scale == 'sanderson':
            return self.calc_en_sanderson()
        else:
            raise ValueError('unknown <scale> value: {}'.format(scale))

    def en_mulliken(self, charge=0, missingIsZero=False, useNegativeEA=False):
        '''
        Return the absolute electronegativity (Mulliken scale), calculated as

        .. math::

           \chi = \\frac{I + A}{2}

        where :math:`I` is the ionization energy and :math:`A` is the electron
        affinity
        '''

        if charge == 0:
            ip = self.ionenergies.get(1, None)
            ea = self.electron_affinity
        elif charge > 0:
            ip = self.ionenergies.get(charge + 1, None)
            ea = self.ionenergies.get(charge, None)
        else:
            raise ValueError('Charge has to be a non-negative integer, got: {}'.format(charge))

        if ip is not None:
            if ea is not None:
                if ea < 0.0 and useNegativeEA:
                    return (ip + ea) * 0.5
                else:
                    return ip * 0.5
            elif ea is None and missingIsZero:
                return ip * 0.5
        else:
            return None

    def en_calc(self, radius='covalent_radius_pyykko', rpow=1, apow=1,
                **zeffkwargs):
        '''
        Calculate the electronegativity from a general formula

        .. math::

           \chi = \left(\\frac{Z_{\\text{eff}}}{r^{\\beta}}\\right)^{\\alpha}

        where

        - :math:`Z_{\\text{eff}}` is the effective nuclear charge
        - :math:`r` is the covalent radius
        - :math:`\\alpha,\\beta` parameters
        '''

        zeff = self.zeff(**zeffkwargs)
        r = getattr(self, radius)

        return math.pow(zeff / math.pow(r, rpow), apow)

    def en_martynov_batsanov(self):
        '''
        Calculates the electronegativity value according to Martynov and
        Batsanov as the average of the ionization energies of the valence
        electrons

        .. math::

           \chi_{MB} = \sqrt{\\frac{1}{n_{v}}\sum^{n_{v}}_{k=1} I_{k}}

        where: :math:`n_{v}` is the number of valence electrons and
        :math:`I_{k}` is the :math:`k` th ionization potential.
        '''

        ionenergies = [self.ionenergies.get(i, None)
                       for i in range(1, self.nvalence(method='simple') + 1)]

        if all(ionenergies):
            return np.sqrt(np.array(ionenergies).mean())
        else:
            return None

    def en_li_xue(self, charge=0, radius='crystal_radius'):
        '''
        Calculate the electronegativity of an atom according to the definition
        of Li and Xue

        Args:
            charge : int
                Charge of the ion
            radius : str
                Type of radius to be used in the calculation, either
                `crystal_radius` as recommended in the paper or `ionic_radius`

        Returns:
            out : dict
                A dictionary with electronegativities as values and
                coordination string as keys or tuple of coordination and spin
                if the ion is LS or HS
        '''

        if charge is None or not isinstance(charge, int) or charge == 0:
            raise ValueError('charge should be a nonzero  initeger')

        neff = {1: 0.85, 2: 1.99, 3: 2.89, 4: 3.45, 5: 3.85, 6: 4.36, 7: 4.99}
        RY = 13.605693009

        if charge == 0:
            Ie = self.electron_affinity
        elif charge > 0:
            Ie = self.ionenergies.get(charge, None)

        crs = [(IR.coordination, IR.spin, getattr(IR, radius))
               for IR in self.ionic_radii if IR.charge == charge]

        out = {}
        for coord, spin, cr in crs:
            # the 100.0 factor converts picometers to Angstroms
            eneg = neff[self.ec.max_n()] * math.sqrt(Ie / RY) * 100.0 / cr
            if len(spin) < 1:
                out[coord] = eneg
            else:
                out[(coord, spin)] = eneg

        return out

    def calc_en_sanderson(self, radius='covalent_radius_pyykko'):
        '''Sanderson electronegativity

        .. math::

           \chi = \\frac{AD}{AD_{\\text{ng}}}

        Args:
            radius : str
                Radius to use in the calcualtion
        '''

        from mendeleev.utils import estimate

        r = getattr(self, radius)
        rng = estimate(self.atomic_number, radius)

        return math.pow(rng / r, 3)

    def nvalence(self, method=None):
        '''Return the number of valence electrons'''

        return self.ec.nvalence(self.block, method=method)

    def __str__(self):
        return "{0} {1} {2}".format(self.atomic_number, self.symbol, self.name)

    def __repr__(self):
        return "%s(\n%s)" % (
            (self.__class__.__name__),
            ' '.join(["\t%s=%r,\n" % (key, getattr(self, key))
                      for key in sorted(self.__dict__.keys())
                      if not key.startswith('_')]))


class IonicRadius(Base):
    '''
    Effective ionic radii and crystal radii in pm retrieved from [1]_.

    .. [1] Shannon, R. D. (1976). Revised effective ionic radii and systematic
       studies of interatomic distances in halides and chalcogenides. Acta
       Crystallographica Section A.
       `doi:10.1107/S0567739476001551 <http://www.dx.doi.org/10.1107/S0567739476001551>`_

    Attributes:
      atomic_number : int
        Atomic number
      charge : int
        Charge of the ion
      econf : str
        Electronic configuration of the ion
      coordination : str
        Type of coordination
      spin : str
        Spin state: HS - high spin, LS - low spin
      crystal_radius : float
        Crystal radius in pm
      ionic_radius : float
        Ionic radius in pm
      origin : str
        Source of the data
      most_reliable : bool
        Most reliable value (see reference)
    '''

    __tablename__ = 'ionicradii'

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey('elements.atomic_number'))
    charge = Column(Integer)
    econf = Column(String)
    coordination = Column(String)
    spin = Column(String)
    crystal_radius = Column(Float)
    ionic_radius = Column(Float)
    origin = Column(String)
    most_reliable = Column(Boolean)

    def __str__(self):
        out = ["{0}={1:>4d}", "{0}={1:5s}", "{0}={1:>6.3f}", "{0}={1:>6.3f}"]
        keys = ['charge', 'coordination', 'crystal_radius', 'ionic_radius']
        return ", ".join([o.format(k, getattr(self, k)) for o, k in zip(out, keys)])

    def __repr__(self):
        return "%s(\n%s)" % (
            (self.__class__.__name__),
            ' '.join(["\t%s=%r,\n" % (key, getattr(self, key))
                      for key in sorted(self.__dict__.keys())
                      if not key.startswith('_')]))


class IonizationEnergy(Base):
    '''
    Ionization energy of an element

    Attributes:
      atomic_number : int
        Atomic number
      degree : int
        Degree of ionization with respect to neutral atom
      energy : float
        Ionization energy in eV parsed from
        http://physics.nist.gov/cgi-bin/ASD/ie.pl on April 13, 2015
    '''

    __tablename__ = 'ionizationenergies'

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey('elements.atomic_number'))
    degree = Column(Integer)
    energy = Column(Float)

    def __str__(self):

        return "{1:5d} {2:10.5f}".format(self.degree, self.energy)

    def __repr__(self):

        return "<IonizationEnergy(atomic_number={a:5d}, degree={d:3d}, energy={e:10.5f})>".format(
            a=self.atomic_number, d=self.degree, e=self.energy)


class OxidationState(Base):
    '''
    Oxidation states of an element

    Attributes:
      atomic_number : int
        Atomic number
      oxidation_state : int
        Oxidation state
    '''

    __tablename__ = 'oxidationstates'

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    oxidation_state = Column(Integer)

    def __repr__(self):

        return "<OxidationState(atomic_number={a:5d}, oxidation_state={o:5d})>".format(
            a=self.atomic_number, o=self.oxidation_state)


class Group(Base):
    '''Name of the group in the periodic table.'''

    __tablename__ = 'groups'

    group_id = Column(Integer, primary_key=True)
    symbol = Column(String)
    name = Column(String)

    def __repr__(self):

        return "<Group(symbol={s:s}, name={n:s})>".format(
            s=self.symbol, n=self.name)


class Series(Base):
    '''
    Name of the series in the periodic table.

    Attributes:
      name : str
        Name of the series
      color : str
        The HEX representation of a color of the series, the colors were
        obtained from
        `ColorBrewer <http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=10>`_
        the qualitative 10-class paired colormap
    '''

    __tablename__ = 'series'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    color = Column(String)

    def __repr__(self):

        return "<Series(name={n:s}, color={c:s})>".format(n=self.name,
                                                          c=self.color)


class Isotope(Base):
    '''
    Isotope

    Attributes:
      abundance : float
        Abundance of the isotope
      atomic_number : int
        Atomic number
      half_life : float
        Half life time
      half_life_unit : str
        Unit for the half life time
      is_radioactive : bool
        A flag marking wheather the isotope is radioactive
      mass : float
        Mass of the isotope
      mass_number : int
        Mass number of the isotope
      mass_uncertainty : float
        Uncertainty of the mass
    '''

    __tablename__ = "isotopes"

    id = Column(Integer, primary_key=True)
    abundance = Column(Float)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    g_factor = Column(Float)
    half_life = Column(Float)
    half_life_unit = Column(String)
    is_radioactive = Column(Boolean)
    mass = Column(Float)
    mass_number = Column(Integer)
    mass_uncertainty = Column(Float)
    spin = Column(Float)
    quadrupole_moment = Column(Float)

    def __str__(self):

        return "{0:5d} {1:5d} {2:10.5f}".format(
            self.atomic_number, self.mass_number, self.mass)

    def __repr__(self):

        return "<Isotope(Z={}, A={}, mass={})>".format(
            self.atomic_number, self.mass_number, self.mass)


class ScreeningConstant(Base):
    '''
    Nuclear screening constants from Clementi, E., & Raimondi, D. L. (1963).
    Atomic Screening Constants from SCF Functions. The Journal of Chemical
    Physics, 38(11), 2686.  `doi:10.1063/1.1733573
    <http://www.dx.doi.org/10.1063/1.1733573>`_ and Clementi, E. (1967). Atomic
    Screening Constants from SCF Functions. II. Atoms with 37 to 86 Electrons.
    The Journal of Chemical Physics, 47(4), 1300.  `doi:10.1063/1.1712084
    <http://www.dx.doi.org/10.1063/1.1712084>`_

    Attributes:
      atomic_number : int
        Atomic number
      n : int
        Principal quantum number
      s : str
        Subshell label, (s, p, d, ...)
      screening : float
        Screening constant
    '''

    __tablename__ = 'screeningconstants'

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    n = Column(Integer)
    s = Column(String)
    screening = Column(Float)

    def __str__(self):

        return "{0:4d} {1:3d} {2:s} {3:10.4f}".format(
            self.atomic_number, self.n, self.s, self.screening)

    def __repr__(self):

        return "<ScreeningConstant(Z={0:4d}, n={1:3d}, s={2:s}, screening={3:10.4f})>".format(
            self.atomic_number, self.n, self.s, self.screening)



# -*- coding: utf-8 -*-

"""
    Setup file for the mendeleev package.
"""

from setuptools import setup, find_packages

# Change these settings according to your needs
MAIN_PACKAGE = "mendeleev"
DESCRIPTION = "Python API with a database of atomic properties for elements in the periodic table"
LICENSE = "MIT"
URL = "https://bitbucket.org/lukaszmentel/mendeleev"
DOWNLOAD_URL = 'https://bitbucket.org/lukaszmentel/mendeleev/get/v0.4.1.tar.gz'
AUTHOR = "Lukasz Mentel"
EMAIL = "lmmentel@gmail.com"
VERSION = '0.4.1'
KEYWORDS = ['periodic', 'table', 'elements', 'atomic', 'properties',
            'mendeleev']


CLASSIFIERS = ['Development Status :: 3 - Alpha',
               'Environment :: Console',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: MIT License',
               'Natural Language :: English',
               'Programming Language :: Python',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 3',
               'Topic :: Scientific/Engineering :: Chemistry',
               'Topic :: Scientific/Engineering :: Physics']


DEPENDENCIES = ['numpy', 'pandas', 'sqlalchemy', 'colorama', 'pyfiglet']


def readme():
    '''Return the contents of the README.rst file.'''
    with open('README.rst') as freadme:
        return freadme.read()


def setup_package():

    setup(name=MAIN_PACKAGE,
          version=VERSION,
          url=URL,
          download_url=DOWNLOAD_URL,
          description=DESCRIPTION,
          author=AUTHOR,
          author_email=EMAIL,
          include_package_data=True,
          install_requires=DEPENDENCIES,
          keywords=KEYWORDS,
          license=LICENSE,
          long_description=readme(),
          classifiers=CLASSIFIERS,
          packages=find_packages(exclude=['tests', 'tests.*']),
          entry_points={
                'console_scripts': [
                    'element.py = mendeleev.cli:clielement',
                    'mdlvappdata = mendeleev.utils:get_app_data']
          },
          )


if __name__ == "__main__":
    setup_package()
